@@ -1,12 +1,16 @@
+import logging
 import math
 import os
 import pathlib
 import re
 import sys
-from hashlib import sha256
+from contextlib import contextmanager
+from functools import partial
+from hashlib import md5
 from urllib.parse import urlsplit
 
 DEFAULT_BLOCK_SIZE = 5 * 2 ** 20
+PY36 = sys.version_info < (3, 7)
 
 
 def infer_storage_options(urlpath, inherit_storage_options=None):
@@ -69,7 +73,7 @@ def infer_storage_options(urlpath, inherit_storage_options=None):
         # https://github.com/dask/dask/issues/1417
         options["host"] = parsed_path.netloc.rsplit("@", 1)[-1].rsplit(":", 1)[0]
 
-        if protocol in ("s3", "gcs", "gs"):
+        if protocol in ("s3", "s3a", "gcs", "gs"):
             options["path"] = options["host"] + options["path"]
         else:
             options["host"] = options["host"]
@@ -96,11 +100,12 @@ def update_storage_options(options, inherited=None):
         inherited = {}
     collisions = set(options) & set(inherited)
     if collisions:
-        collisions = "\n".join("- %r" % k for k in collisions)
-        raise KeyError(
-            "Collision between inferred and specified storage "
-            "options:\n%s" % collisions
-        )
+        for collision in collisions:
+            if options.get(collision) != inherited.get(collision):
+                raise KeyError(
+                    "Collision between inferred and specified storage "
+                    "option:\n%s" % collision
+                )
     options.update(inherited)
 
 
@@ -115,7 +120,7 @@ def infer_compression(filename):
     extension. This includes builtin (gz, bz2, zip) compressions, as well as
     optional compressions. See fsspec.compression.register_compression.
     """
-    extension = os.path.splitext(filename)[-1].strip(".")
+    extension = os.path.splitext(filename)[-1].strip(".").lower()
     if extension in compressions:
         return compressions[extension]
 
@@ -276,7 +281,7 @@ def tokenize(*args, **kwargs):
     """
     if kwargs:
         args += (kwargs,)
-    return sha256(str(args).encode()).hexdigest()
+    return md5(str(args).encode()).hexdigest()
 
 
 def stringify_path(filepath):
@@ -301,11 +306,16 @@ def stringify_path(filepath):
     Any other object is passed through unchanged, which includes bytes,
     strings, buffers, or anything else that's not even path-like.
     """
-    if hasattr(filepath, "__fspath__"):
+    if isinstance(filepath, str):
+        return filepath
+    elif hasattr(filepath, "__fspath__"):
         return filepath.__fspath__()
     elif isinstance(filepath, pathlib.Path):
         return str(filepath)
-    return filepath
+    elif hasattr(filepath, "path"):
+        return filepath.path
+    else:
+        return filepath
 
 
 def make_instance(cls, args, kwargs):
@@ -327,7 +337,7 @@ def common_prefix(paths):
     return "/".join(parts[0][:i])
 
 
-def other_paths(paths, path2, is_dir=None):
+def other_paths(paths, path2, is_dir=None, exists=False):
     """In bulk file operations, construct a new file tree from a list of files
 
     Parameters
@@ -341,6 +351,9 @@ def other_paths(paths, path2, is_dir=None):
         For the special case where the input in one element, whether to regard the value
         as the target path, or as a directory to put a file path within. If None, a
         directory is inferred if the path ends in '/'
+    exists: bool (optional)
+        For a str destination, it is already exists (and is a dir), files should
+        end up inside.
 
     Returns
     -------
@@ -351,6 +364,8 @@ def other_paths(paths, path2, is_dir=None):
         path2 = path2.rstrip("/")
         if len(paths) > 1:
             cp = common_prefix(paths)
+            if exists:
+                cp = cp.rsplit("/", 1)[0]
             path2 = [p.replace(cp, path2, 1) for p in paths]
         else:
             if is_dir:
@@ -374,7 +389,7 @@ def get_protocol(url):
 
 
 def can_be_local(path):
-    """Can the given URL be used wih open_local?"""
+    """Can the given URL be used with open_local?"""
     from fsspec import get_filesystem_class
 
     try:
@@ -384,23 +399,6 @@ def can_be_local(path):
         return False
 
 
-def setup_logger(logname, level="DEBUG", clear=True):
-    """Add standard logging handler to logger of given name"""
-    import logging
-
-    logger = logging.getLogger(logname)
-    if clear:
-        logger.handlers.clear()
-    handle = logging.StreamHandler()
-    formatter = logging.Formatter(
-        "%(asctime)s - %(name)s - %(levelname)s " "- %(message)s"
-    )
-    handle.setFormatter(formatter)
-    logger.addHandler(handle)
-    logger.setLevel(level)
-    return logger
-
-
 def get_package_version_without_import(name):
     """For given package name, try to find the version without importing it
 
@@ -435,3 +433,103 @@ def get_package_version_without_import(name):
         return mod.__version__
     except (ImportError, AttributeError):
         return None
+
+
+def setup_logging(logger=None, logger_name=None, level="DEBUG", clear=True):
+    if logger is None and logger_name is None:
+        raise ValueError("Provide either logger object or logger name")
+    logger = logger or logging.getLogger(logger_name)
+    handle = logging.StreamHandler()
+    formatter = logging.Formatter(
+        "%(asctime)s - %(name)s - %(levelname)s - %(funcName)s -- %(message)s"
+    )
+    handle.setFormatter(formatter)
+    if clear:
+        logger.handlers.clear()
+    logger.addHandler(handle)
+    logger.setLevel(level)
+    return logger
+
+
+def _unstrip_protocol(name, fs):
+    if isinstance(fs.protocol, str):
+        if name.startswith(fs.protocol):
+            return name
+        return fs.protocol + "://" + name
+    else:
+        if name.startswith(tuple(fs.protocol)):
+            return name
+        return fs.protocol[0] + "://" + name
+
+
+def mirror_from(origin_name, methods):
+    """Mirror attributes and methods from the given
+    origin_name attribute of the instance to the
+    decorated class"""
+
+    def origin_getter(method, self):
+        origin = getattr(self, origin_name)
+        return getattr(origin, method)
+
+    def wrapper(cls):
+        for method in methods:
+            wrapped_method = partial(origin_getter, method)
+            setattr(cls, method, property(wrapped_method))
+        return cls
+
+    return wrapper
+
+
+@contextmanager
+def nullcontext(obj):
+    yield obj
+
+
+def merge_offset_ranges(paths, starts, ends, max_gap=0, max_block=None, sort=True):
+    """Merge adjacent byte-offset ranges when the inter-range
+    gap is <= `max_gap`, and when the merged byte range does not
+    exceed `max_block` (if specified). By default, this function
+    will re-order the input paths and byte ranges to ensure sorted
+    order. If the user can guarantee that the inputs are already
+    sorted, passing `sort=False` will skip the re-ordering.
+    """
+
+    # Check input
+    if not isinstance(paths, list):
+        raise TypeError
+    if not isinstance(starts, list):
+        starts = [starts] * len(paths)
+    if not isinstance(ends, list):
+        ends = [starts] * len(paths)
+    if len(starts) != len(paths) or len(ends) != len(paths):
+        raise ValueError
+
+    # Sort by paths and then ranges if `sort=True`
+    if sort:
+        paths, starts, ends = [list(v) for v in zip(*sorted(zip(paths, starts, ends)))]
+
+    if paths:
+        # Loop through the coupled `paths`, `starts`, and
+        # `ends`, and merge adjacent blocks when appropriate
+        new_paths = paths[:1]
+        new_starts = starts[:1]
+        new_ends = ends[:1]
+        for i in range(1, len(paths)):
+            if (
+                paths[i] != paths[i - 1]
+                or ((starts[i] - new_ends[-1]) > max_gap)
+                or ((max_block is not None and (ends[i] - new_starts[-1]) > max_block))
+            ):
+                # Cannot merge with previous block.
+                # Add new `paths`, `starts`, and `ends` elements
+                new_paths.append(paths[i])
+                new_starts.append(starts[i])
+                new_ends.append(ends[i])
+            else:
+                # Merge with previous block by updating the
+                # last element of `ends`
+                new_ends[-1] = ends[i]
+        return new_paths, new_starts, new_ends
+
+    # `paths` is empty. Just return input lists
+    return paths, starts, ends
